# coding: utf-8

import json
import pycrunch
from pycrunch.shoji import TaskError

from scrunch.helpers import shoji_view_wrapper, shoji_entity_wrapper


class ScriptExecutionError(Exception):
    def __init__(self, client_error, resolutions):
        self.client_error = client_error
        self.resolutions = resolutions

    def __repr__(self):
        return json.dumps(self.resolutions, indent=2)


DEFAULT_SUBVARIABLE_SYNTAX = False


class BaseScript:
    def __init__(self, resource):
        """
        :param resource: Pycrunch Entity.
        """
        self.resource = resource

    def get_default_syntax_flag(self, strict_subvariable_syntax):
        """
        The Syntax flag tells scripts whether to be strict about
        evaluating subvariable codes with the array[subvar] square bracket
        syntax (when True) or to allow referring to subvariables by their
        global code only  (when False).

        If the user did provide a flag (not None), honor it and send that
        in the request.

        IFF the user did NOT provide a flag (is None), then see if the
        session's feature flags know something that's been provided from the
        server, and if nothing's there use the constant's default (currently
        False).

        :param strict_subvariable_syntax: User provided flag
        :return: bool
        """
        if strict_subvariable_syntax is not None:
            return strict_subvariable_syntax
        flags = self.resource.session.feature_flags
        return flags.get("clients_strict_subvariable_syntax", DEFAULT_SUBVARIABLE_SYNTAX)

    def execute(self, script_body, strict_subvariable_syntax=None):
        pass


class SystemScript(BaseScript):

    def format_request_url(self, request_url, strict_subvariable_syntax=None):
        strict_subvariable_syntax_flag = self.get_default_syntax_flag(strict_subvariable_syntax)
        if strict_subvariable_syntax_flag:
            request_url += "?strict_subvariable_syntax=true"
        return request_url

    def execute(self, script_body, strict_subvariable_syntax=None):
        """
        Will run a system script on this.

        System scripts do not have a return value. If they execute correctly
        they'll finish silently. Otherwise, an error will raise.
        """
        # The script execution endpoint is a shoji:view
        payload = shoji_view_wrapper(script_body)
        try:
            execute_url = self.format_request_url(self.resource.views['execute'], strict_subvariable_syntax)
            return self.resource.session.post(execute_url, json=payload)
        except pycrunch.ClientError as err:
            resolutions = err.args[2]["resolutions"]
            raise ScriptExecutionError(err, resolutions)


class DatasetScripts(BaseScript):

    def execute(self, script_body, strict_subvariable_syntax=None, dry_run=False):
        strict_subvariable_syntax = self.get_default_syntax_flag(strict_subvariable_syntax)
        payload = {
            "body": script_body,
            "strict_subvariable_syntax": strict_subvariable_syntax
        }

        if dry_run:
            payload["dry_run"] = True
            method = self.resource.scripts.post
        else:
            method = self.resource.scripts.create

        # The dataset script execution endpoint is a shoji:entity
        body = shoji_entity_wrapper(payload)

        try:
            method(body)
        except pycrunch.ClientError as err:
            if isinstance(err, TaskError):
                # For async script validation error
                resolutions = err.message["resolutions"]
                raise ScriptExecutionError(err, resolutions)
            elif err.status_code == 400:
                # For sync validation
                resolutions = err.args[2]["resolutions"]
                raise ScriptExecutionError(err, resolutions)
            raise err  # 404 or something else

    def dry_run(self, script_body, strict_subvariable_syntax=None):
        self.execute(script_body, strict_subvariable_syntax, dry_run=True)

    def collapse(self):
        """
        When a dataset has too many scripts. Collapse will concatenate
        all the previously executed scripts into one the first. It will delete
        all savepoints associated with the collapsed scripts.
        """
        self.resource.scripts.collapse.post({})

    def all(self):
        scripts_index = self.resource.scripts.index
        scripts = []
        for s_url, s in scripts_index.items():
            scripts.append(s.entity)
        scripts = sorted(scripts, key=lambda s: s.body["creation_time"])
        return scripts

    def revert_to(self, id=None, script_number=None):
        all_scripts = self.all()
        if script_number is not None:
            script = all_scripts[script_number]
        elif id is not None:
            # We have to do this because currently the API does not expose the
            # script ID directly.
            script = [s for s in all_scripts if "scripts/{}/".format(id) in s.self][0]
        else:
            raise ValueError("Must indicate either ID or script number")

        resp = script.revert.post({})  # Asynchronous request
        pycrunch.shoji.wait_progress(resp, self.resource.session)



# coding: utf-8

import uuid
import pytest

from pycrunch.shoji import as_entity

from fixtures import BaseIntegrationTestCase
from scrunch.accounts import Account


class TestAccount(BaseIntegrationTestCase):

    def test_current_account(self):
        act = Account.current_account(self.site)
        assert act.resource.self == self.site.account.self

    def test_execute(self):
        raise pytest.skip("Wait until account execute is implemented in backend")

    def test_account_projects(self):
        # Create a top level project
        project_name = uuid.uuid4().hex
        api_project = self.site.projects.create(as_entity({"name": project_name}))
        api_project.refresh()
        act = Account.current_account(self.site)

        project = act.projects.by_name(project_name)
        assert project.url == api_project.self
        assert project.name == project_name

        project = act.projects.by_id(api_project.body["id"])
        assert project.url == api_project.self

        project.delete()



# coding: utf-8

from pycrunch.shoji import as_entity

from scrunch.mutable_dataset import get_mutable_dataset
from fixtures import BaseIntegrationTestCase


class TestViews(BaseIntegrationTestCase):
    FIXTURE_VARIABLES = [
        ("var1", "numeric"),
        ("var2", "text"),
        ("var3", "numeric"),
    ]

    def _create_ds(self):
        ds = self.site.datasets.create(as_entity({"name": "test_script"})).refresh()

        for alias, v_type in self.FIXTURE_VARIABLES:
            var_body = as_entity({
                "name": alias,
                "alias": alias,
                "type": v_type,
            })
            ds.variables.create(var_body)

        return ds

    def test_create_view(self):
        api_ds = self._create_ds()
        scrunch_dataset = get_mutable_dataset(api_ds.body.id, self.site, editor=True)
        new_view = scrunch_dataset.views.create("My first view", None)
        # Assert it is a view
        assert new_view.resource.body["view_of"] == api_ds.self

        # Assert the variables are here
        for alias, v_type in self.FIXTURE_VARIABLES:
            assert new_view[alias].type == v_type

    def test_create_view_non_personal_ds(self):
        api_ds = self._create_ds()

        # Put the dataset in a public project
        project = self.site.projects.create(as_entity({"name": "public project"}))
        project.patch({"index": {api_ds.self: {}}})

        scrunch_dataset = get_mutable_dataset(api_ds.body.id, self.site, editor=True)
        new_view = scrunch_dataset.views.create("My view", None)

        # Assert it is a view
        assert new_view.resource.body["view_of"] == api_ds.self
        assert new_view.resource.body["owner"] == project.self

        # Assert the variables are here
        for alias, v_type in self.FIXTURE_VARIABLES:
            assert new_view[alias].type == v_type

    def test_create_var_subset(self):
        api_ds = self._create_ds()
        scrunch_dataset = get_mutable_dataset(api_ds.body.id, self.site, editor=True)
        new_view = scrunch_dataset.views.create("Sub view", ["var1", "var2"])
        # Assert it is a view
        assert new_view.resource.body["view_of"] == api_ds.self

        # Only the desired variables are part of the view
        assert set(new_view.keys()) == {"var1", "var2"}

    def test_fetch_view_names(self):
        api_ds = self._create_ds()
        scrunch_dataset = get_mutable_dataset(api_ds.body.id, self.site, editor=True)
        view1 = scrunch_dataset.views.create("view 1", None)
        view2 = scrunch_dataset.views.create("view 2", ["var1", "var2"])

        assert scrunch_dataset.views.all_names() == {view1.name, "view 2"}

        fetched_view = scrunch_dataset.views.get_by_name("view 2")
        assert fetched_view.id == view2.id




# coding: utf-8

import six
import json
from mock import Mock
from requests import Response
from unittest import TestCase

from pycrunch.shoji import Entity, Catalog, Order

from scrunch.order import InvalidPathError
from scrunch.scripts import ScriptExecutionError
from scrunch.datasets import Project, ProjectDatasetsOrder, get_personal_project

from .mock_session import MockSession


class TestProjectNesting(TestCase):
    def test_detect_correct_handler(self):
        session = Mock(
            feature_flags={'old_projects_order': True}
        )
        dataset_order = Order(session, **{
            'graph': []
        })
        datasets_catalog = Catalog(session, **{
            'index': {},
            'order': dataset_order
        })
        shoji_resource = Entity(session, **{
            'self': '/project/url/',
            'body': {},
            'index': {},
            'datasets': datasets_catalog
        })
        project = Project(shoji_resource)
        self.assertTrue(isinstance(project.order, ProjectDatasetsOrder))

        session = Mock(
            feature_flags={'old_projects_order': False}
        )
        shoji_resource = Entity(session, **{
            'self': '/project/url/',
            'body': {},
            'index': {},
            'graph': []
        })
        project = Project(shoji_resource)
        self.assertTrue(isinstance(project.order, Project))

    def test_create_subproject(self):
        session = MockSession()
        session.feature_flags = {'old_projects_order': False}
        shoji_resource = Entity(session, **{
            'self': 'http://example.com/project/url/',
            'body': {},
            'index': {},
            'graph': [],
        })

        # Setup the POST request and the fixture for the GET that happens after
        # the .refresh()
        created_project_url = 'http://example.com/project/2/'
        response = Response()
        response.status_code = 201
        response.headers = {
            'Location': created_project_url
        }
        session.add_post_response(response)
        session.add_fixture(created_project_url, {
            'self': created_project_url,
            'body': {},
            'index': {},
            'graph': [],
        })
        project = Project(shoji_resource)

        # Create a new project
        pa = project.order.create_project("Project A")
        self.assertTrue(isinstance(pa, Project))

        # Check that we sent the correct payload to the server
        self.assertEqual(pa.url, created_project_url)
        post_request = session.requests[-2]
        refresh_request = session.requests[-1]
        self.assertEqual(refresh_request.method, 'GET')
        self.assertEqual(post_request.method, 'POST')
        self.assertEqual(post_request.url, project.url)
        self.assertEqual(json.loads(post_request.body), {
            'element': 'shoji:entity',
            'body': {
                'name': 'Project A'
            }
        })

    def make_tree(self):
        session = MockSession()
        session.feature_flags = {'old_projects_order': False}

        #       A
        #     /   \
        #    B     C
        #    |
        #    D
        projects_res_url = 'http://example.com/api/projects/'
        a_res_url = 'http://example.com/api/projects/A/'
        b_res_url = 'http://example.com/api/projects/B/'
        c_res_url = 'http://example.com/api/projects/C/'
        d_res_url = 'http://example.com/api/projects/D/'
        a_payload = {
            'element': 'shoji:entity',
            'self': a_res_url,
            'catalogs': {
                'project': 'http://example.com/api/projects/'
            },
            'body': {
                'name': 'project A'
            },
            'index': {
                b_res_url: {
                    'id': 'idB',
                    'name': 'project B',
                    'icon': None,
                    'description': '',
                    'type': 'project'
                },
                c_res_url: {
                    'id': 'idC',
                    'name': 'project C',
                    'icon': None,
                    'description': '',
                    'type': 'project'
                }
            },
            'graph': [c_res_url, b_res_url]
        }
        b_payload = {
            'element': 'shoji:entity',
            'self': b_res_url,
            'catalogs': {
                'project': a_res_url
            },
            'body': {'name': 'project B'},
            'index': {
                d_res_url: {
                    'id': 'idD',
                    'name': 'project D',
                    'icon': None,
                    'description': '',
                    'type': 'project'
                }
            },
            'graph': [d_res_url]
        }
        c_payload = {
            'element': 'shoji:entity',
            'self': c_res_url,
            'catalogs': {
                'project': a_res_url
            },
            'body': {'name': 'project C'},
            'index': {},
            'graph': []
        }
        d_payload = {
            'element': 'shoji:entity',
            'self': d_res_url,
            'catalogs': {
                'project': b_res_url
            },
            'body': {'name': 'project D'},
            'index': {},
            'graph': []
        }
        projects_catalog = {
            'element': 'shoji:catalog',
            'self': projects_res_url,
            'index': {
                a_res_url: {
                    'name': 'project A'
                }
            }
        }
        session.add_fixture(a_res_url, a_payload)
        session.add_fixture(b_res_url, b_payload)
        session.add_fixture(c_res_url, c_payload)
        session.add_fixture(d_res_url, d_payload)
        session.add_fixture(projects_res_url, projects_catalog)

        response_204 = Response()
        response_204.status_code = 204
        session.add_patch_response(response_204)
        session.add_patch_response(response_204)
        return session

    def test_follow_path(self):
        a_res_url = 'http://example.com/api/projects/A/'
        d_res_url = 'http://example.com/api/projects/D/'

        session = self.make_tree()
        a_res = session.get(a_res_url).payload
        project_a = Project(a_res)
        project_c = project_a.order['| project C ']
        project_d = project_a.order['| project B | project D']
        self.assertTrue(isinstance(project_d, Project))
        self.assertEqual(project_d.resource.self, d_res_url)

        with self.assertRaises(InvalidPathError):
            project_a.order['| project B | Invalid']

    def test_rename(self):
        a_res_url = 'http://example.com/api/projects/A/'
        session = self.make_tree()
        a_res = session.get(a_res_url).payload
        project_a = Project(a_res)
        project_d = project_a.order['| project B | project D']
        project_d.rename('Renamed Project D')
        # This works because .rename() implementation calls shoji Entity.edit
        # which will make the request an update the resource's internal payload
        # as well. If this passes it means that Scrunch is correct and pycrunch
        # did its thing.
        self.assertEqual(project_d.resource.body.name, 'Renamed Project D')
        self.assertEqual(project_d.name, 'Renamed Project D')

    def test_move_things(self):
        a_res_url = 'http://example.com/api/projects/A/'
        dataset_url = 'http://example.com/api/datasets/1/'
        session = self.make_tree()
        project_a = Project(session.get(a_res_url).payload)
        project_c = project_a.order['| project C ']
        project_d = project_a.order['| project B | project D']
        dataset = Mock(url=dataset_url)

        # Moving project C under project D
        project_d.move_here([project_c, dataset])

        # After a move_here there is a PATCH and a GET and a GET for each item
        # the PATCH performs the changes and the GET is a resource.refresh()
        # The last request was made on the `dataset` variable which is a Mock
        # so doesn't register a request
        dataset.resource.refresh.assert_called_once()
        patch_request = session.requests[-3]
        refresh_request = session.requests[-2]
        self.assertEqual(refresh_request.method, 'GET')
        self.assertEqual(refresh_request.url, project_d.url)
        self.assertEqual(patch_request.method, 'PATCH')
        self.assertEqual(patch_request.url, project_d.url)
        self.assertEqual(json.loads(patch_request.body), {
            'element': 'shoji:entity',
            'body': {},
            'index': {
                project_c.url: {},
                dataset.url: {}
            },
            # 'graph': [project_c.url, dataset.url]
        })

    def test_move_project(self):
        catalog_url = 'http://example.com/api/projects/'
        a_res_url = 'http://example.com/api/projects/A/'
        d_res_url = 'http://example.com/api/projects/D/'
        session = self.make_tree()
        project_a = Project(session.get(a_res_url).payload)
        project_d = Project(session.get(d_res_url).payload)

        with self.assertRaises(InvalidPathError):
            # Not a root path, invalid.
            project_d.move("%s" % project_a.name)

        # Moving to A
        project_d.move("| %s" % project_a.name)

        # Will have to iteratively nagivate the path making requests to
        # projects root and then to A
        request1 = session.requests[-5]
        request2 = session.requests[-4]
        self.assertEqual(request1.url, catalog_url)
        self.assertEqual(request2.url, a_res_url)

        patch_request = session.requests[-3]
        self.assertEqual(patch_request.method, 'PATCH')
        self.assertEqual(patch_request.url, project_a.url)
        index = json.loads(patch_request.body)['index']
        self.assertEqual(index, {
            project_d.url: {},
        })

    def test_place(self):
        a_res_url = 'http://example.com/api/projects/A/'
        dataset1_url = 'http://example.com/api/datasets/1/'
        dataset2_url = 'http://example.com/api/datasets/2/'
        session = self.make_tree()
        project_a = Project(session.get(a_res_url).payload)
        project_b = project_a.order['| project B']
        project_d = project_a.order['| project B | project D']
        dataset1 = Mock(url=dataset1_url)
        dataset2 = Mock(url=dataset2_url)
        dataset1.name = 'Dataset 1'
        dataset2.name = 'Dataset 2'

        # Do a .place call
        project_a.place(dataset1, '| project B', before='project D')

        # After a move_here there is a PATCH and a GET
        # the PATCH performs the changes and the GET is a resource.refresh()
        patch_request = session.requests[-2]
        self.assertEqual(patch_request.method, 'PATCH')

        # Note the patch is to project B even though we did `.place` on
        # project A, but the target path pointed to B
        self.assertEqual(patch_request.url, project_b.url)
        self.assertEqual(json.loads(patch_request.body), {
            'element': 'shoji:entity',
            'body': {},
            'index': {
                dataset1.url: {}
            },
            # Note how the graph sent includes dataset1.url before project D
            'graph': [dataset1.url, project_d.url]
        })

        # Since the PATCH did not really update the server or the session
        # test fixtures, we need to update the fixtures to reflect the fact
        # that they've been modified by the recent PATCH request
        session.adapter.fixtures[project_b.url]['index'][dataset1.url] = {
            'name': dataset1.name,
            'type': 'dataset'
        }
        session.adapter.fixtures[project_b.url]['graph'] = [dataset1.url, project_d.url]
        project_a.place(dataset2, '| project B', after='Dataset 1')
        patch_request = session.requests[-2]
        self.assertEqual(patch_request.method, 'PATCH')
        self.assertEqual(patch_request.url, project_b.url)
        self.assertEqual(json.loads(patch_request.body), {
            'element': 'shoji:entity',
            'body': {},
            'index': {
                dataset2.url: {}
            },
            # Dataset 2 got placed after dataset 1 :)
            'graph': [dataset1.url, dataset2.url, project_d.url]
        })

    def test_reorder(self):
        session = self.make_tree()
        a_res_url = 'http://example.com/api/projects/A/'
        b_res_url = 'http://example.com/api/projects/B/'
        c_res_url = 'http://example.com/api/projects/C/'
        project_a = Project(session.get(a_res_url).payload)
        project_a.reorder(["project C", "project B"])
        patch_request = session.requests[-2]
        self.assertEqual(patch_request.method, 'PATCH')
        self.assertEqual(patch_request.url, project_a.url)
        self.assertEqual(json.loads(patch_request.body), {
            'element': 'shoji:entity',
            'body': {},
            'index': {},
            'graph': [c_res_url, b_res_url]
        })

    def test_move(self):
        session = self.make_tree()
        a_res_url = 'http://example.com/api/projects/A/'
        b_res_url = 'http://example.com/api/projects/B/'
        c_res_url = 'http://example.com/api/projects/C/'
        project_a = Project(session.get(a_res_url).payload)
        project_a.reorder(["project C", "project B"])
        patch_request = session.requests[-2]
        self.assertEqual(patch_request.method, 'PATCH')
        self.assertEqual(patch_request.url, project_a.url)
        self.assertEqual(json.loads(patch_request.body), {
            'element': 'shoji:entity',
            'body': {},
            'index': {},
            'graph': [c_res_url, b_res_url]
        })

    def test_is_root(self):
        a_res_url = 'http://example.com/api/projects/A/'
        session = self.make_tree()
        project_a = Project(session.get(a_res_url).payload)
        project_b = project_a.order['| project B']
        self.assertTrue(project_a.is_root)
        self.assertFalse(project_b.is_root)

    def test_children(self):
        session = self.make_tree()
        a_res_url = 'http://example.com/api/projects/A/'
        project_a = Project(session.get(a_res_url).payload)
        # Get instantiated correctly
        self.assertTrue(all(isinstance(c, Project) for c in project_a.children))
        # Get iterated on the right order
        self.assertEqual([c.url for c in project_a.children],
                         project_a.resource.graph)
        # First request before .child should .refresh() the resource to
        # ensure fresh data
        refresh_request = session.requests[0]
        self.assertEqual(refresh_request.method, 'GET')
        self.assertEqual(refresh_request.url, a_res_url)

    def test_delete_project(self):
        mock_resource = Mock()
        project = Project(mock_resource)
        project.delete()
        mock_resource.delete.assert_called_once()


class TestPersonalProject(TestCase):
    PERSONAL_URL = "http://example.com/api/projects/personal/"

    def _get_root(self):
        session = MockSession()
        root_url = "http://example.com/api/"
        projects_url = "http://example.com/api/projects/"
        session.add_fixture(root_url, {
            "self": root_url,
            "element": "shoji:catalog",
            "catalogs": {
                "projects": projects_url
            },
            "index": {}
        })
        session.add_fixture(projects_url, {
            "self": projects_url,
            "element": "shoji:catalog",
            "catalogs": {
                "personal": self.PERSONAL_URL
            },
            "index": {}
        })
        session.add_fixture(self.PERSONAL_URL, {
            "self": self.PERSONAL_URL,
            "element": "shoji:catalog",
            "index": {}
        })
        response = Response()
        response.status_code = 204
        session.add_patch_response(response)
        return session.get(root_url).payload

    def test_get_personal(self):
        root = self._get_root()
        personal = get_personal_project(root)
        self.assertEqual(personal.url, self.PERSONAL_URL)

    def test_move_to_personal(self):
        root = self._get_root()
        a_res_url = 'http://example.com/api/projects/A/'
        root.session.add_fixture(a_res_url, {
            'self': a_res_url,
            'element': 'shoji:entity',
            'catalogs': {
                'project': 'http://example.com/api/projects/'
            },
            'body': {
                'name': 'project A'
            },
            'index': {},
            'graph': []
        })
        personal = get_personal_project(root)
        project_a = Project(root.session.get(a_res_url).payload)
        personal.move_here(project_a)
        patch_request = root.session.requests[-3]
        self_refresh_request = root.session.requests[-2]
        child_refresh_request = root.session.requests[-1]
        self.assertEqual(child_refresh_request.method, 'GET')
        self.assertEqual(self_refresh_request.method, 'GET')
        self.assertEqual(self_refresh_request.url, self.PERSONAL_URL)
        self.assertEqual(patch_request.url, self.PERSONAL_URL)
        self.assertEqual(patch_request.method, 'PATCH')
        self.assertEqual(json.loads(patch_request.body)['index'], {
            project_a.url: {}
        })


class TestProjectScripts(TestCase):
    project_execute_url = "http://example.com/project/id/execute/"

    def test_running_script(self):
        session = MockSession()
        shoji_resource = Entity(session, **{
            'self': 'http://example.com/project/id/',
            'body': {},
            'index': {},
            'graph': [],
            "views": {
                "execute": self.project_execute_url
            }
        })

        exexute_resource = Catalog(session, **{
            'self': self.project_execute_url,
        })

        response = Response()
        response.status_code = 204
        session.add_post_response(response)
        session.add_fixture(self.project_execute_url, exexute_resource)
        project = Project(shoji_resource)

        # Execute script on this project
        script_body = "NOOP;"
        project.execute(script_body)

        # Verify the POST request was sent to the correct url with entity payload
        execution_request = session.requests[-1]
        assert execution_request.method == "POST"
        assert execution_request.url == self.project_execute_url
        assert json.loads(execution_request.body) == {
            'element': 'shoji:view',
            'value': script_body,
        }

    def test_error_handling(self):
        session = MockSession()
        shoji_resource = Entity(session, **{
            'self': 'http://example.com/project/id/',
            'body': {},
            'index': {},
            'graph': [],
            "views": {
                "execute": self.project_execute_url
            }
        })

        execute_resource = Catalog(session, **{
            'self': self.project_execute_url
        })

        resolutions = [{"line": 100}]
        resolutions_out = json.dumps({"resolutions": resolutions})
        if six.PY3:
            # Python3 needs a bytes object, not str
            resolutions_out = resolutions_out.encode("utf-8")
        error_response = Response()
        error_response.status_code = 400
        error_response.headers = {"Content-Type": "application/json"}
        error_response._content = resolutions_out
        error_response.request = Mock(url=self.project_execute_url)
        session.add_post_response(error_response)

        session.add_fixture(self.project_execute_url, execute_resource)
        project = Project(shoji_resource)

        # Script will raise exception
        with self.assertRaises(ScriptExecutionError) as err:
            project.execute("Bad script")
        assert err.exception.resolutions == [{"line": 100}]
        assert err.exception.client_error.status_code == 400




# coding: utf-8

import json
import pytest
from mock import Mock, MagicMock
from requests import Response

from pycrunch.shoji import Entity, Catalog, Order

from .mock_session import MockSession
from scrunch.views import DatasetViews
from scrunch.mutable_dataset import MutableDataset


class TestViews:
    def test_disabled_on_views(self):
        session = Mock()
        dataset_shoji_resource = Entity(session, **{
            'self': '/datasets/id/',
            'body': {
                "view_of": None
            },
        })
        view_shoji_resource = Entity(session, **{
            'self': '/datasets/id/',
            'body': {
                "view_of": "some dataset URL"
            },
        })
        for res in (dataset_shoji_resource, view_shoji_resource):
            res.variables = MagicMock()
            res.folders = MagicMock()

        dataset = MutableDataset(dataset_shoji_resource)
        view = MutableDataset(view_shoji_resource)
        assert not dataset.is_view
        assert view.is_view

        with pytest.raises(TypeError) as err:
            view.views  # Accessing this raises on views
        assert str(err.value) == "Cannot access views of views"

        views_handler = dataset.views
        assert isinstance(views_handler, DatasetViews)

    folders_url = "http://host/folders/url"
    public_url = "http://host/folders/public/url"
    project_url = "http://host/project/url/"
    variables_url = "http://host/variables/url/"
    root_url = "http://host/api/"
    datasets_url = "http://host/datasets/url/"

    def _set_folders_fixtures(self, session):
        folders_catalog_res = Catalog(session, **{
            "self": self.folders_url,
            "index": {},
            "catalogs": {
                "public": self.public_url
            },
        })
        public_catalog_res = Catalog(session, **{
            "self": self.public_url,
            "index": {},
            "body": {
                "name": "folder name"
            },
            "catalogs": {},
        })
        project_catalog_res = Catalog(session, **{
            "self": self.project_url,
            "index": {},
            "body": {},
        })
        root_catalog_res = Catalog(session, **{
            "self": self.root_url,
            "index": {},
            "catalogs": {"datasets": self.datasets_url}
        })
        dataset_catalog_res = Catalog(session, **{
            "self": self.datasets_url,
            "index": {},
        })
        session.site_url = self.root_url
        session.add_fixture(self.folders_url, folders_catalog_res)
        session.add_fixture(self.public_url, public_catalog_res)
        session.add_fixture(self.project_url, project_catalog_res)
        session.add_fixture(self.root_url, root_catalog_res)
        session.add_fixture(self.datasets_url, dataset_catalog_res)

    def test_create_view_no_cols(self):
        session = MockSession()
        views_url = "http://host/views/url/"
        dataset_url = "http://host/datasets/id/"
        variables_url = "http://host/variables/url/"
        new_view_url = "http://host/new/view/url/"
        project_url = "http://host/project/id"
        ds_res = Entity(session, **{
            'self': dataset_url,
            'body': {
                "view_of": None,
                "owner": project_url
            },
            "catalogs": {
                "views": views_url,
                "variables": variables_url,
                "parent": self.project_url
            }
        })
        variables_catalog_res = Catalog(session, **{
            "self": self.variables_url,
            "index": {}
        })
        new_view_res = Entity(session, **{
            "self": new_view_url,
            "body": {
                "view_of": dataset_url
            },
            "catalogs": {
                "variables": variables_url,  # Same as dataset, unimportant
                "folders": self.folders_url
            }
        })
        self._set_folders_fixtures(session)
        session.add_fixture(self.variables_url, variables_catalog_res)

        create_response = Response()
        create_response.status_code = 201
        create_response.headers = {
            "Location": new_view_url
        }
        session.add_post_response(create_response)
        session.add_fixture(new_view_url, new_view_res)

        views = DatasetViews(ds_res)
        new_view = views.create("My view", None)
        assert isinstance(new_view, MutableDataset)
        assert new_view.is_view
        assert new_view.resource.body["view_of"] == dataset_url
        create_body = json.loads(session.requests[2].body)
        assert create_body == {
            "element": "shoji:entity",
            "body": {
                "name": "My view",
                "view_of": dataset_url,
                "owner": project_url,
            },
        }

    def test_create_view_with_cols(self):
        session = MockSession()
        views_url = "http://host/views/url/"
        dataset_url = "http://host/datasets/id/"
        new_view_url = "http://host/new/view/url/"
        project_url = "http://host/project/id"
        ds_res = Entity(session, **{
            'self': dataset_url,
            'body': {
                "view_of": None,
                "owner": project_url,
            },
            "catalogs": {
                "views": views_url,
                "variables": self.variables_url,
                "parent": self.project_url
            }
        })
        variables_catalog_res = Catalog(session, **{
            "self": self.variables_url,
            "index": {
                "idA": {"alias": "A"},
                "idB": {"alias": "B"},
            }
        })
        new_view_res = Entity(session, **{
            "self": new_view_url,
            "body": {
                "view_of": dataset_url
            },
            "catalogs": {
                "variables": self.variables_url,  # Same as dataset, unimportant
                "folders": self.folders_url
            }
        })
        self._set_folders_fixtures(session)
        session.add_fixture(self.variables_url, variables_catalog_res)

        create_response = Response()
        create_response.status_code = 201
        create_response.headers = {
            "Location": new_view_url
        }
        session.add_post_response(create_response)
        session.add_fixture(new_view_url, new_view_res)

        views = DatasetViews(ds_res)
        views.create("My view", ["A", "B"])
        create_body = json.loads(session.requests[3].body)
        assert create_body == {
            "element": "shoji:entity",
            "body": {
                "view_cols": ["idA", "idB"],
                "name": "My view",
                "view_of": dataset_url,
                "owner": project_url
            },
        }

    def test_all_names(self):
        session = MockSession()
        views_url = "http://host/views/url/"
        dataset_url = "http://host/datasets/id/"
        ds_res = Entity(session, **{
            'self': dataset_url,
            'body': {
                "view_of": None
            },
            "catalogs": {
                "views": views_url,
                "variables": self.variables_url,
                "parent": self.project_url
            }
        })
        views_catalog = Catalog(session, **{
            "self": views_url,
            "index": {
                "v1_url": {"name": "view 1"},
                "v2_url": {"name": "view 2"},
            }
        })
        session.add_fixture(views_url, views_catalog)
        views = DatasetViews(ds_res)
        view_names = views.all_names()
        assert view_names == {"view 1", "view 2"}

    def test_get_by_name(self):
        session = MockSession()
        views_url = "http://host/views/url/"
        dataset_url = "http://host/datasets/id/"
        variables_url = "http://host/variables/url/"
        view_1_url = "http://host/view/1/"
        ds_res = Entity(session, **{
            'self': dataset_url,
            'body': {
                "view_of": None
            },
            "catalogs": {
                "views": views_url,
                "variables": variables_url,
                "parent": self.project_url
            }
        })
        views_catalog = Catalog(session, **{
            "self": views_url,
            "index": {
                view_1_url: {"name": "view 1"},
                "v2_url": {"name": "view 2"},
            }
        })

        view_1_res = Entity(session, **{
            "self": view_1_url,
            "body": {
                "view_of": dataset_url
            },
            "catalogs": {
                "variables": variables_url,  # Same as dataset, unimportant
                "folders": self.folders_url
            }
        })
        self._set_folders_fixtures(session)
        session.add_fixture(self.variables_url, {"index": {}})
        session.add_fixture(views_url, views_catalog)
        session.add_fixture(view_1_url, view_1_res)
        views = DatasetViews(ds_res)
        view_1 = views.get_by_name("view 1")
        assert view_1.url == view_1_url
